﻿/***************************************************************************
 * [ソフトウェア名]
 *      NeuralNetwork.cs
 * [開発者]
 *      K.Morishita (Kumamoto-Univ. @ 2011.12)
 * [概要]
 *      ニューラルネットワークを利用した識別器のクラスです。
 *      ソースコードの改造が容易になるように、とことんオブジェクトに分解しました。
 *      おかげで学習は遅くなったようですが、可読性と可塑性が増したのではないかと思います。
 *      ニューラルネットに関してはシロートが作っていますので変なところがあるかもしれません。
 *      XORを学習できることは確認済みです。
 * [参考資料]
 *      浅川伸一, バックプロパゲーション, http://www.cis.twcu.ac.jp/~asakawa/waseda2002/bp.pdf ,2011/12. （少し誤字があるけど）
 * [検討課題]
 *      特徴の評価のための機構組み込み（ジャックナイフは実装が必要だろうなぁ）
 * [履歴]
 *      2011/12/8   開発開始
 *                  コンセプトは、使い易く書きやすく。。。
 *                  効率的で高速な演算とは縁遠い構造をしています。
 *                  オブジェクトや演算の関係が分かり易ければ良いという考えの基で作りました。
 *      2011/12/9   基本動作の確認が行えるまでは組めた。
 *                  今後の予定：学習終了トリガとなる構造体の宣言
 *                              学習モードの整備
 *                              保存されたファイルから文字列を解析して、NeuralNetTeacherクラスの再構成・NNクラスの再構成を実施するメソッドを整備
 *      2011/12/11  昨日に引き続き、デバッグとコメント見直しを行った。
 *      2011/12/24  昨日より再度整備を開始
 *                  ニューラルネットの核はほぼできたと思う。ただし、関数近似に使用できない。どうすればよいのだろう？
 *                  今日は学習用のクラスを整備中
 *      2011/12/25  学習用のクラスはファイル保存機能を除いてほぼ完全に整備できたと思う。
 *                  次は保存メソッドと、識別機能利用時のファイルからの復元機能を設計する必要がある。
 *      2011/12/27  昨日の内に保存ファイルからの復帰機能を実装できた。
 *                  今日はUIの改善と動作テストを行った。
 *                  本日で基本的な機能の実装は終わったと思う。
 *                  無論、学習の方法や評価の方法に思うところはあるのだけどこれで一つの区切りとする。
 *      2011/12/30  結局気になっていたので学習器の学習アルゴリズムを、モデル毎にまとめて学習する方法から均等に学習する（引数によってモデル毎にもできる）方法に変えた。
 *                  それに伴いモデルクラスも改良した。
 *                  結果はというと…鳥の学習ではとりわけよくなった。
 *                  鳥だと、特定の種だけ2倍学習させただけで途端に成績が悪くなる。
 *      2012/3/10   GetIDandLikelihoodにおいて、入力された出力層の値が非値だった場合にエラーが発生していたのを訂正
 * *************************************************************************/
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;

namespace PatternRecognition
{
    /// <summary>
    /// ベクトルクラス
    /// </summary>
    public class Vector
    { 
        /************** メンバ変数 **********************/
        /// <summary>ベクトル要素</summary>
        private double[] vector;
        /************** プロパティ **********************/
        /// <summary>
        /// ベクトルの長さ
        /// </summary>
        public int Length { get { return this.vector.Length; } }
        /************** インデクサ **********************/
        /// <summary>
        /// インデクサ
        /// <para>配列と同様にアクセスできます。</para>
        /// </summary>
        /// <param name="index">要素番号</param>
        /// <returns>指定されたベクトル要素を返す</returns>
        public double this[int index]
        {
            get { if (index >= 0 && index < this.vector.Length) return this.vector[index]; else throw new SystemException("範囲外の要素が指定されました。"); }
            set { if (index >= 0 && index < this.vector.Length) this.vector[index] = value; else throw new SystemException("範囲外の要素が指定されました。"); }
        }
        /************** メソッド **********************/
        /// <summary>
        /// ベクトル要素を文字列として返す
        /// <para>出力フォーマット付</para>
        /// </summary>
        /// <param name="format">出力フォーマット</param>
        /// <returns>ベクトル要素の中身</returns>
        public string ToString(string format = "E4")
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder(300);
            for (int i = 0; i < this.vector.Length; i++) sb.Append(this.vector[i].ToString(format)).Append(",");
            return sb.ToString();
        }
        /// <summary>
        /// 特徴ベクトルを追加する
        /// <para>後追加となる</para>
        /// </summary>
        /// <param name="addtionalVector">追加したいベクトル</param>
        public void Add(double[] addtionalVector)
        { 
            int size = this.Length;
            Array.Resize(ref this.vector, this.vector.Length + addtionalVector.Length);
            for(int i = size, k = 0; i < this.Length; i++, k++)
                this.vector[i] = addtionalVector[k];

            return;
        }
        /// <summary>
        /// ベクトル情報をdouble型配列で返す
        /// <para>シャローコピーと同じなので取扱注意のこと。不用意にいじると、ベクトルの中身が変わります。。</para>
        /// </summary>
        /// <returns>配列表現の特徴ベクトル</returns>
        public double[] ToDoubleArray()
        {
            return this.vector;
        }
        #region /* コンストラクタ関係 -----------------------------------------------------------------*/
        /// <summary>
        /// double型配列を使った初期化付コンストラクタ
        /// </summary>
        /// <param name="_vector">double型配列</param>
        /// <exception cref="SystemException">配列がnullならスロー</exception>
        public Vector(double[] _vector)
        {
            if (_vector == null) throw new SystemException("配列がnullです。");
            this.vector = _vector;
        }
        /// <summary>
        /// 文字列を利用したコンストラクタ
        /// <para>カンマ・タブ・半角スペースで区切られた文字列を渡して下さい。</para>
        /// <para>解析に成功すると文字列に含まれていた数値がセットされます。</para>
        /// </summary>
        /// <param name="_vector">文字列表現のベクトル成分</param>
        public Vector(string _vector)
        {
            this.vector = new double[0];
            string[] field = _vector.Split(',');                        // 文字列を分割処理
            if (field.Length == 1) field = _vector.Split('\t');
            if (field.Length == 1) field = _vector.Split(' ');
            double value;
            for (int i = 0; i < field.Length; i++)                      // 文字列を解析して配列要素に格納していく
            {
                Boolean success = double.TryParse(field[i], out value);
                if (success)
                {
                    Array.Resize(ref this.vector, this.vector.Length + 1);
                    this.vector[this.vector.Length - 1] = value;
                }
            }
            return;
        }
        /// <summary>
        /// サイズ指定のコンストラクタ
        /// <para>指定したサイズのベクトルを生成します。生成後に一つ一つの要素にアクセスする用途を想定しています。</para>
        /// </summary>
        /// <param name="size">ベクトルのサイズ</param>
        public Vector(int size)
        { 
            this.vector = new double[size];
        }
        /// <summary>
        /// 要素数1のベクトルを生成します
        /// <para>関数近似などの用途を想定しています。</para>
        /// </summary>
        /// <param name="value">格納したい数</param>
        public Vector(double value)
        {
            this.vector = new double[1];
            this.vector[0] = value;
        }
        /// <summary>
        /// 初期化なしのコンストラクタ
        /// <para>空のベクトルを生成します。適宜要素を追加するような用途での使用を想定しています。</para>
        /// </summary>
        public Vector() 
        {
            this.vector = new double[0];
        }
        #endregion
    }
    /// <summary>
    /// 特徴ベクトル
    /// <para>Vectorクラスのラッパー</para>
    /// <para>クラス名から特徴ベクトルであることを連想しやすいようにラップした。</para>
    /// </summary>
    public class Feature : Vector
    {
        #region /* コンストラクタ関係 -----------------------------------------------------------------*/
        /// <summary>
        /// double型配列を使った初期化付コンストラクタ
        /// </summary>
        /// <param name="_vector">double型配列による特徴ベクトル</param>
        /// <exception cref="SystemException">配列がnullならスロー</exception>
        public Feature(double[] _vector)
            : base(_vector)
        {
        }
        /// <summary>
        /// 文字列を利用したコンストラクタ
        /// <para>カンマ・タブ・半角スペースで区切られた文字列を渡して下さい。</para>
        /// <para>解析に成功すると文字列に含まれていた数値がセットされます。</para>
        /// </summary>
        /// <param name="_vector">文字列で表現した特徴ベクトル</param>
        public Feature(string _vector)
            : base(_vector)
        {
        }
        /// <summary>
        /// サイズ指定のコンストラクタ
        /// <para>指定したサイズのベクトルを生成します。生成後に一つ一つの要素にアクセスする用途を想定しています。</para>
        /// </summary>
        /// <param name="size">ベクトルサイズ</param>
        public Feature(int size)
            : base(size)
        { 
        }
        /// <summary>
        /// 要素数1のベクトルを生成します
        /// <para>関数近似などの用途を想定しています。</para>
        /// </summary>
        /// <param name="value">格納したい数</param>
        public Feature(double value)
            : base(value)
        {
        }
        /// <summary>
        /// 初期化なしのコンストラクタ
        /// <para>空のベクトルを生成します。適宜要素を追加するような用途での使用を想定しています。</para>
        /// </summary>
        public Feature()
            : base()
        {
        }
        #endregion
    }
    /// <summary>
    /// ニューラルネットワーク処理クラス
    /// </summary>
    public partial class NeuralNetwork
    {
        /// <summary>
        /// ニューラルネットのユニット
        /// </summary>
        private class Unit
        {
            /* メンバ変数 ***********************************/
            /// <summary>前の層との間の重みベクトル</summary>
            private Vector weightVector;
            /// <summary>前の層との間の重みベクトルの修正量</summary>
            private Vector correctionWeightVector;
            /// <summary>入力値</summary>
            private double inputValue;
            /// <summary>出力値</summary>
            private double outputValue;
            /// <summary>出力誤差</summary>
            private double error;
            /// <summary>発火の閾値</summary>
            private double threshold;
            /// <summary>乱数ジェネレータ</summary>
            static private Random myRandom;
            /* プロパティ ***********************************/
            /// <summary>
            /// 重みベクトルサイズ
            /// </summary>
            public int Length { get { return this.weightVector.Length; } }
            /// <summary>
            /// 本ユニットの出力
            /// </summary>
            public double Output { get { return this.outputValue; } set { this.outputValue = value; } }
            /// <summary>
            /// 本ユニットの誤差
            /// </summary>
            public double Error
            {
                get { return this.error; }
                set { this.error = value; }
            }
            /// <summary>
            /// 発火の閾値
            /// </summary>
            public double Threshold
            {
                get { return this.threshold; }
                set { this.threshold = value; }
            }
            /* メソッド ***********************************/
            /// <summary>
            /// 重み行列を文字列化する
            /// </summary>
            /// <returns>文字列化した重み行列</returns>
            public override string ToString()
            {
                System.Text.StringBuilder sb = new System.Text.StringBuilder(5000);
                sb.Append(this.weightVector.ToString());
                return sb.ToString();
            }
            /// <summary>
            /// 指定された修正量で指定された結合係数を修正する
            /// </summary>
            /// <param name="amountOfCorrection">修正量</param>
            /// <param name="index">インデックス番号</param>
            public void CalibrateWeightVector(double amountOfCorrection, int index)
            {
                this.weightVector[index] += amountOfCorrection;
                return;
            }
            /// <summary>
            /// 結合係数の修正量をセットする
            /// <para>本メソッドを呼び出しただけでは修正されません。DoCorrect()を呼び出すことで結合係数を修正します。</para>
            /// </summary>
            /// <param name="correction"></param>
            /// <param name="index"></param>
            public void SetCorrectionWeight(double correction, int index)
            {
                this.correctionWeightVector[index] = correction;
                return;
            }
            /// <summary>
            /// 結合係数を予めセットしている修正量で修正します
            /// <para>「新しい結合係数 = 古い結合係数 + 修正量」で定義しています。</para>
            /// </summary>
            public void DoCorrect()
            {
                for (int i = 0; i < this.weightVector.Length; i++) this.weightVector[i] += this.correctionWeightVector[i];
                return;
            }
            /// <summary>
            /// 出力値を計算してメンバ変数に保存する
            /// </summary>
            private void CalcOutput()
            {
                this.outputValue = 1.0 / (1 + Math.Exp(-this.inputValue - this.threshold));
                //this.outputValue = 2.0 / (1 + Math.Exp(-this.inputValue - this.threshold)) - 1.0;
            }
            /// <summary>
            /// 本ユニットに信号を入力し、出力を返す
            /// </summary>
            /// <param name="inputSignal">入力信号（ニューラルネットなら、前の層の出力ベクトル）</param>
            /// <returns>本ユニットの出力</returns>
            public double GetOutput(Vector inputSignal)
            {
                if (this.weightVector.Length != inputSignal.Length) throw new SystemException("重みベクトルの次元数と引数の次元数が異なります。");
                this.outputValue = 0.0;
                this.inputValue = 0.0;
                for (int i = 0; i < inputSignal.Length; i++) this.inputValue += this.weightVector[i] * inputSignal[i];
                this.CalcOutput();
                return this.outputValue;
            }
            /// <summary>
            /// 入力値を加算する
            /// <para>同時に、出力も計算する</para>
            /// </summary>
            /// <param name="value"></param>
            public void AddInputValue(double value)
            {
                this.inputValue += value;
                this.CalcOutput();
                return;
            }
            /// <summary>
            /// 発火閾値を初期化する
            /// <para>発火閾値は0.0～1.0の間の乱数で初期化されます。</para>
            /// </summary>
            public void ClearThreshold()
            {
                this.threshold = this.GetRandom();
                return;
            }
            /// <summary>
            /// 入力値を初期化する
            /// <para>入力値は0.0で初期化されます。</para>
            /// </summary>
            public void ClearInputValue()
            {
                this.inputValue = this.GetRandom();
                return;
            }
            /// <summary>
            /// 重みベクトルをセット（上書き）する
            /// </summary>
            /// <param name="_weightVector">セットする重みベクトル</param>
            public void SetWeightVector(double[] _weightVector)
            {
                if (this.weightVector.Length != _weightVector.Length) throw new SystemException("重みベクトルの大きさが一致しません。");
                for (int i = 0; i < this.weightVector.Length; i++) this.weightVector[i] = _weightVector[i];
                return;
            }
            /// <summary>
            /// 重みベクトルをセット（上書き）する
            /// </summary>
            /// <param name="_weightVector">セットする重みベクトル</param>
            public void SetWeightVector(Vector _weightVector)
            {
                if (this.weightVector.Length != _weightVector.Length) throw new SystemException("重みベクトルの大きさが一致しません。");
                for (int i = 0; i < this.weightVector.Length; i++) this.weightVector[i] = _weightVector[i];
                return;
            }
            /// <summary>
            /// 乱数を返す
            /// </summary>
            /// <returns>発生させた乱数</returns>
            private double GetRandom()
            {
                double v = 4.0 * myRandom.NextDouble() - 2.0;
                if (Math.Abs(v) < 0.1) v = this.GetRandom();    // 0近くだと学習が進まないため、バイアスを設ける
                return v;
            }
            /// <summary>
            /// 重みベクトルを乱数で初期化する
            /// </summary>
            public void InitWeightVectorWithRandom()
            {
                for (int i = 0; i < this.weightVector.Length; i++) this.weightVector[i] = this.GetRandom();
                return;
            }
            /// <summary>
            /// 指定要素の重みを返す
            /// </summary>
            /// <param name="index">要素番号</param>
            /// <returns>指定された要素番号の重み</returns>
            /// <exception cref="SystemException">指定要素番号が範囲内を超えるとスロー</exception>
            public double GetWeight(int index)
            {
                if (index >= 0 && index < this.weightVector.Length)
                    return this.weightVector[index];
                else
                    throw new SystemException("要素が範囲内を超えました。");
            }
            /// <summary>
            /// 重みベクトルを返す
            /// </summary>
            /// <returns>重みベクトル</returns>
            public Vector GetWeightVector()
            {
                return this.weightVector;
            }
            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="preLayerSize">前段のユニット数</param>
            public Unit(int preLayerSize)
            {
                if (myRandom == null) myRandom = new Random();
                this.weightVector = new Vector(preLayerSize);
                this.correctionWeightVector = new Vector(this.weightVector.Length);
                this.InitWeightVectorWithRandom();
                this.ClearInputValue();
                this.ClearThreshold();
                this.Error = 0.0;
            }
        }
    } // ニューラルネットのユニット
    public partial class NeuralNetwork
    {
        /// <summary>
        /// ニューラルネットワークのパラメータ
        /// <para>主にパラメータの受け渡しと設定値の不正チェックを目的とする。</para>
        /// </summary>
        public struct Parameter
        {
            /************** 定数 **********************/
            /// <summary>入力層のユニット数がこの文字列に続きます</summary>
            private const string keywordOfUnitInputLayer = "Number of unit in input-layer=";    // 検索・保存に利用する
            /// <summary>中間層のユニット数がこの文字列に続きます</summary>
            private const string keywordOfUnitHiddenLayer = "Number of unit in hidden-layer=";
            /// <summary>出力層のユニット数がこの文字列に続きます</summary>
            private const string keywordOfUnitOutputLayer = "Number of unit in output-layer=";
            /// <summary>中間層数がこの文字列に続きます</summary>
            private const string keywordOfNumHiddenLayer = "Number of hidden-layer=";
            /// <summary>学習係数がこの文字列に続きます</summary>
            private const string keywordOfLearningCoefficient = "Learning coefficient=";
            /************** メンバ変数 **********************/
            /// <summary>入力層のユニット数</summary>
            private readonly int numOfUnitInInputLayer;
            /// <summary>中間層のユニット数</summary>
            private readonly int numOfUnitInHiddenLayer;
            /// <summary>出力層のユニット数</summary>
            private readonly int numOfUnitInOutputLayer;
            /// <summary>中間層数</summary>
            private readonly int numOfHiddenLayer;
            /// <summary>学習係数</summary>
            private readonly double learningCoefficient;
            /************** プロパティ **********************/
            /// <summary>
            /// 入力層のユニット数
            /// </summary>
            public int NumOfUnitInInputLayer { get { return this.numOfUnitInInputLayer; } }
            /// <summary>
            /// 中間層のユニット数
            /// </summary>
            public int NumOfUnitInHiddenLayer { get { return this.numOfUnitInHiddenLayer; } }
            /// <summary>
            /// 出力層のユニット数
            /// </summary>
            public int NumOfUnitInOutputLayer { get { return this.numOfUnitInOutputLayer; } }
            /// <summary>
            /// 合計層数
            /// </summary>
            public int TotalLayer { get { return this.numOfHiddenLayer + 2; } }
            /// <summary>
            /// 学習係数
            /// </summary>
            public double LearningCoefficient { get { return this.learningCoefficient; } }
            /************** メソッド **********************/
            /// <summary>
            /// パラメータを文字列として返す
            /// </summary>
            /// <returns>文字列化したパラメータ</returns>
            public override string ToString()
            {
                System.Text.StringBuilder sb = new System.Text.StringBuilder(1500);
                sb.Append("<Neural Network Parameter>").Append("\n");
                sb.Append(keywordOfUnitInputLayer).Append(this.NumOfUnitInInputLayer.ToString()).Append("\n");
                sb.Append(keywordOfUnitHiddenLayer).Append(this.NumOfUnitInHiddenLayer.ToString()).Append("\n");
                sb.Append(keywordOfUnitOutputLayer).Append(this.NumOfUnitInOutputLayer.ToString()).Append("\n");
                sb.Append(keywordOfNumHiddenLayer).Append(this.numOfHiddenLayer.ToString()).Append("\n");
                sb.Append(keywordOfLearningCoefficient).Append(this.LearningCoefficient.ToString("0.00#")).Append("\n");
                sb.Append("</Neural Network Parameter>").Append("\n");
                return sb.ToString();
            }
            /// <summary>
            /// パラメータのチェックを行う
            /// <para>エラーがあれば例外をスロー</para>
            /// </summary>
            /// <param name="_numOfUnitInInputLayer">入力層のユニット数</param>
            /// <param name="_numOfUnitInHiddenLayer">中間層のユニット数</param>
            /// <param name="_numOfUnitInOutputLayer">出力層のユニット数</param>
            /// <param name="_numOfHiddenLayer">中間層数</param>
            /// <param name="_learningCoefficient">学習係数</param>
            private static void CheckARGV(int _numOfUnitInInputLayer, int _numOfUnitInHiddenLayer, int _numOfUnitInOutputLayer, int _numOfHiddenLayer, double _learningCoefficient)
            {
                if (_numOfUnitInInputLayer == 0 ||      // 各層のユニット数が0はあり得ない。
                    _numOfUnitInHiddenLayer == 0 ||
                    _numOfUnitInOutputLayer == 0 ||
                    _numOfUnitInHiddenLayer < 0 ||      // 中間層数が0は在り得ても負値は在り得ない
                    _learningCoefficient == 0.0)        // 学習係数0もあり得ない
                    throw new SystemException("構造体へのパラメータ設定値が不正です。");
                return;
            }
            /// <summary>
            /// ファイル名を指定して、パラメータをファイルから読み込ませるコンストラクタ
            /// </summary>
            /// <param name="fname">ファイル名<para>デフォルト："Neural Network Parameter.txt"</para></param>
            public Parameter(string fname = "Neural Network Parameter.txt")
            {
                int _numOfUnitInInputLayer = 0, _numOfUnitInHiddenLayer = 0, _numOfUnitInOutputLayer = 0, _numOfHiddenLayer = 0;
                double _learningCoefficient = 0.0;

                if (System.IO.File.Exists(fname) == false) throw new SystemException("指定されたファイルは存在しません。");
                using (System.IO.StreamReader sr = new System.IO.StreamReader(fname, System.Text.Encoding.UTF8))
                {
                    while (sr.EndOfStream == false)
                    {
                        string line = sr.ReadLine();                            // 一行読み込み
                        Regex ri = new Regex(@"\d+");
                        Regex rd = new Regex(@"\d+.\d+");
                        Match m = ri.Match(line);
                        int i = 0;
                        if(m.ToString() != "") i = int.Parse(m.ToString());
                        m = rd.Match(line);
                        double d = 0.0;
                        if(m.ToString() != "") d = double.Parse(m.ToString());
                        if (line.IndexOf(keywordOfUnitInputLayer) >= 0) _numOfUnitInInputLayer = i;
                        if (line.IndexOf(keywordOfUnitHiddenLayer) >= 0) _numOfUnitInHiddenLayer = i;
                        if (line.IndexOf(keywordOfUnitOutputLayer) >= 0) _numOfUnitInOutputLayer = i;
                        if (line.IndexOf(keywordOfNumHiddenLayer) >= 0) _numOfHiddenLayer = i;
                        if (line.IndexOf(keywordOfLearningCoefficient) >= 0) _learningCoefficient = d;
                    }
                }
                CheckARGV(_numOfUnitInInputLayer, _numOfUnitInHiddenLayer, _numOfUnitInOutputLayer, _numOfHiddenLayer, _learningCoefficient);
                this.numOfUnitInInputLayer = _numOfUnitInInputLayer;
                this.numOfUnitInHiddenLayer = _numOfUnitInHiddenLayer;
                this.numOfUnitInOutputLayer = _numOfUnitInOutputLayer;
                this.numOfHiddenLayer = _numOfHiddenLayer;
                this.learningCoefficient = _learningCoefficient;
            }
            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="_numOfUnitInInputLayer">入力層のユニット数<para>特徴ベクトルの次元数と同じ数を指定して下さい。</para></param>
            /// <param name="_numOfUnitInHiddenLayer">中間層のユニット数<para>この数が多いと、過学習に陥りやすくなります。n個で2^n状態を表すことができることを考慮して決定して下さい。</para></param>
            /// <param name="_numOfUnitInOutputLayer">出力層のユニット数<para>この数は出力状態量と同じ数を指定して下さい。</para></param>
            /// <param name="_numOfHiddenLayer">中間層数<para>この数+1だけ識別面（超平面）が形成されます。特徴量の分布をよく検討しなければまずいことになりそうです。</para></param>
            /// <param name="_learningCoefficient">学習係数<para>この数が大きいと結合係数の修正量が大きくなりすぎていつまでも収束しなくなります。かといって小さいといつまでも終わりません。カットアンドトライで確認してみてください。</para></param>
            public Parameter(int _numOfUnitInInputLayer, int _numOfUnitInHiddenLayer, int _numOfUnitInOutputLayer, int _numOfHiddenLayer, double _learningCoefficient)
            {
                CheckARGV(_numOfUnitInInputLayer, _numOfUnitInHiddenLayer, _numOfUnitInOutputLayer, _numOfHiddenLayer, _learningCoefficient);
                if (_numOfUnitInHiddenLayer == 0) Console.WriteLine("中間層数が0でしたので中間層のユニット数は無視しました。");
                this.numOfUnitInInputLayer = _numOfUnitInInputLayer;
                this.numOfUnitInHiddenLayer = _numOfUnitInHiddenLayer;
                this.numOfUnitInOutputLayer = _numOfUnitInOutputLayer;
                this.numOfHiddenLayer = _numOfHiddenLayer;
                this.learningCoefficient = _learningCoefficient;
                return;
            }
        }
    } // ニューラルネットワークのパラメータ
    public partial class NeuralNetwork
    {
        /// <summary>
        /// 入力・中間・出力層クラス
        /// <para>層を配列として宣言して扱ってもよかったのだろうけど、なんだか嫌だったので再帰構造を使ってこのクラスを一つ宣言すれば完結するようにした。</para>
        /// <para>2011/12/11現在、ユニット間の結合係数は層の出力を得るたびに逐次処理を行っているので厳密にはBPのアルゴリズムとはずれるんだけど上手く動いている。</para>
        /// </summary>
        private class Layer
        {
            /* 列挙体 ***********************************/
            /// <summary>
            /// 層の種類
            /// </summary>
            private enum Kind
            {
                /// <summary>入力層</summary>
                InputLayer,
                /// <summary>中間層（隠れ層）</summary>
                HiddenLayer,
                /// <summary>出力層</summary>
                OutputLayer,
                /// <summary>不定</summary>
                NA
            }
            /* メンバ変数 ***********************************/
            /// <summary>総層数</summary>
            private readonly int grossLayerNum;
            /// <summary>通し層番号</summary>
            private readonly int myLayerNum;
            /// <summary>この層が保持するユニット</summary>
            private Unit[] units;
            /// <summary>ニューラルネットのパラメータ</summary>
            private readonly Parameter parameter;
            /// <summary>下層レイヤ</summary>
            private Layer subsidiaryLayer;
            /// <summary>学習係数</summary>
            public static double learningCoefficient = 0.01;
            /// <summary>前層のユニット数</summary>
            private int preLayerUnitNum;
            /// <summary>出力層における平均出力誤差の変化量</summary>
            static private double variationError = 0.0;
            /// <summary>出力層における平均出力誤差のバックアップ</summary>
            static private double backupError = 0.0;
            /// <summary>
            /// 出力層における出力誤差の総計
            /// <para>以下で定義しています。</para>
            /// <para>sumError += Math.Abs(this.units[i].Error);</para>
            /// </summary>
            static private double totalOutputError = 0.0;
            /// <summary>乱数生成器</summary>
            static private Random myRandom;
            /* プロパティ ***********************************/
            /// <summary>
            /// この層の種類
            /// </summary>
            private Kind MyKind
            {
                get
                {
                    if (this.myLayerNum == this.grossLayerNum - 1)
                        return Kind.InputLayer;
                    else if (this.myLayerNum == 0)
                        return Kind.OutputLayer;
                    else
                        return Kind.HiddenLayer;
                }
            }
            /// <summary>
            /// 学習係数
            /// </summary>
            public double LearningCoefficient
            {
                get { return learningCoefficient; }
                set { learningCoefficient = value; }
            }
            /// <summary>
            /// ユニット数
            /// </summary>
            public int Lenght { get { return this.units.Length; } }
            /// <summary>
            /// 出力誤差の変化量
            /// </summary>
            public double VariationOfError { get { return variationError; } }
            /// <summary>
            /// 出力層の全ユニット出力の誤差を合計したもの
            /// <para>絶対値をとった各ユニットの誤差を合計したものです。</para>
            /// </summary>
            public double TotalOutputError { get { return totalOutputError; } }
            /* メソッド ***********************************/
            /// <summary>
            /// この層が保持する全ユニットの結合係数を文字列化する
            /// </summary>
            /// <returns>文字列化したレイヤ情報</returns>
            public override string ToString()
            {
                System.Text.StringBuilder sb = new System.Text.StringBuilder(5000);
                if (this.MyKind != Kind.InputLayer)                                         // 入力層は結合係数を持たないため飛ばす
                {
                    sb.Append("<Layer = ").Append(this.myLayerNum.ToString()).Append(">\n");
                    if (this.MyKind == Kind.HiddenLayer)
                        sb.Append("<Layer kind>Hidden</Layer kind>").Append("\n");
                    else if (this.MyKind == Kind.OutputLayer)
                        sb.Append("<Layer kind>Output</Layer kind>").Append("\n");
                    for (int i = 0; i < this.units.Length; i++)
                    {
                        sb.Append("<unit num=").Append(i.ToString()).Append(">").Append("\n");
                        sb.Append("<threshold>").Append("\n");
                        sb.Append(this.units[i].Threshold.ToString()).Append("\n");
                        sb.Append("</threshold>").Append("\n");
                        sb.Append("<weight vector>").Append("\n");
                        sb.Append(this.units[i].ToString()).Append("\n");
                        sb.Append("</weight vector>").Append("\n");
                        sb.Append("</unit num=").Append(i.ToString()).Append(">").Append("\n");
                    }
                    sb.Append("</Layer = ").Append(this.myLayerNum.ToString()).Append(">\n");
                }
                if (this.MyKind != Kind.OutputLayer) sb.Append(this.subsidiaryLayer.ToString());// 出力層はさらに続く層がない。そのため、入力層と中間層のみ次の層を呼び出している。
                return sb.ToString();
            }
            /// <summary>
            /// 各ユニットが持つ発火の閾値を配列で取得する
            /// </summary>
            /// <returns></returns>
            private double[] GetThreshold()
            {
                double[] threshold = new double[this.units.Length];
                for (int i = 0; i < this.units.Length; i++)
                    threshold[i] = this.units[i].Threshold;
                return threshold;
            }
            /// <summary>
            /// 指定された前段のユニット番号に対する重みを配列で返す
            /// <remarks>
            /// 例えば本オブジェクトが出力層である時に8という数字が指定された場合、直前の中間層の持つ第9ユニットに対する結合係数を返す。
            /// </remarks>
            /// </summary>
            /// <param name="index">前段のユニット番号</param>
            /// <returns>指定ユニットと結びついた重み</returns>
            public double[] GetWeight(int index)
            {
                double[] weight = new double[this.units.Length];
                for (int i = 0; i < this.units.Length; i++) weight[i] = this.units[i].GetWeight(index);
                return weight;
            }
            /// <summary>
            /// 各ユニットの出力誤差を配列で返す
            /// </summary>
            /// <returns>出力誤差</returns>
            public double[] GetError()
            {
                double[] error = new double[this.units.Length];
                for (int i = 0; i < error.Length; i++) error[i] = this.units[i].Error;
                return error;
            }
            /// <summary>
            /// 各ユニットの出力値を配列で返す
            /// </summary>
            /// <returns>ユニットの出力</returns>
            public double[] GetOutput()
            {
                double[] output = new double[this.units.Length];
                for (int i = 0; i < output.Length; i++) output[i] = this.units[i].Output;
                return output;
            }
            /// <summary>
            /// ユニットの出力誤差を計算し、ユニットに記憶させる
            /// </summary>
            /// <param name="NeuralNetTeacher">教師ベクトル</param>
            /// <param name="myOutput">出力ベクトル</param>
            public void CalcOwnUnitsError(Vector NeuralNetTeacher, Vector myOutput)
            {
                if (this.MyKind == Kind.OutputLayer)// 出力層の誤差を計算する-----------------
                {
                    double sumError = 0.0;
                    for (int i = 0; i < this.units.Length; i++)
                    {
                        this.units[i].Error = myOutput[i] - NeuralNetTeacher[i];    // ユニット毎に誤差量を計算する。
                        sumError += Math.Abs(this.units[i].Error);
                        this.units[i].Error += 0.1 * myRandom.NextDouble();         // 乱数をすこーし加えると収束が速い。
                    }
                    totalOutputError = sumError;                                          // 出力層の出力値は学習の収束条件に利用できるので保存しておく
                    double averageError = sumError / (double)this.units.Length;
                    variationError = (averageError - backupError) / backupError;
                    backupError = averageError;
                }
                else if (this.MyKind == Kind.HiddenLayer)// 中間層の誤差を計算する------------
                {
                    double[] postError = this.subsidiaryLayer.GetError();           // 下の層の誤差を取得
                    double[] postOutput = this.subsidiaryLayer.GetOutput();
                    int postUnitSize = this.subsidiaryLayer.Lenght;                 // 下の層のユニット数を変数に確保（配列の判定文に入れると最適化されない限り何度も参照されるので、そのオーバーヘッドを避けるため）
                    for (int i = 0; i < this.units.Length; i++)                     // この層の誤差を計算
                    {
                        double[] postWeight = this.subsidiaryLayer.GetWeight(i);
                        double error = 0.0;
                        for (int j = 0; j < postUnitSize; j++)
                            error += postError[j] * postWeight[j];// * postOutput[j] * (1 - postOutput[j])
                        this.units[i].Error = error + 0.1 * myRandom.NextDouble();
                    }
                }
                return;
            }
            /// <summary>
            /// 結合荷重の修正量を返す
            /// </summary>
            /// <param name="unitOutputError">ユニットの出力誤差</param>
            /// <param name="unitOutput">ユニットの出力値</param>
            /// <param name="preLayerOutput">調整したい結合荷重が接続された前段ユニットの出力値</param>
            /// <returns>重み修正量</returns>
            public double GetAmountOfCorrection(double unitOutputError, double unitOutput, double preLayerOutput)
            {
                return -learningCoefficient * unitOutputError * 2.0 * unitOutput * (1 - unitOutput) * preLayerOutput;
            }
            /// <summary>
            /// 結合係数修正量の計算、および発火の閾値修正を実施する
            /// <para>ユニットの誤差量を計算した後に呼び出してください。</para>
            /// </summary>
            /// <param name="preLayerOutput">前段の層の出力</param>
            /// <param name="myOutput">この層の出力ベクトル</param>
            /// <exception cref="SystemException">入力層で本メソッドが呼び出されるとスロー</exception>
            private void CalibrateWeightAndThreshold(Vector preLayerOutput, Vector myOutput)
            {
                if (this.MyKind != Kind.InputLayer)                                 // 出力層や中間層なら処理する
                {
                    // この層が保持するユニット全てに対して処理を実施する
                    for (int i = 0; i < this.units.Length; i++)
                    {
                        // 前段のユニット数だけ結合係数がある。その分ループ
                        for (int j = 0; j < this.preLayerUnitNum; j++)
                        {
                            double amountOfCorrection = GetAmountOfCorrection(this.units[i].Error, myOutput[i], preLayerOutput[j]);
                            this.units[i].SetCorrectionWeight(amountOfCorrection, j);
                        }
                        // バイアスの修正
                        this.units[i].Threshold += GetAmountOfCorrection(this.units[i].Error, myOutput[i], 1.0);
                    }
                }
                else
                    throw new SystemException("この層は入力層なので結合係数の修正は必要ありません。");
                return;
            }
            /// <summary>
            /// 全てのユニットに対して予めセットした修正量を適用して、結合係数を調整する
            /// <para>より下層に位置するLayerの修正も指示する。</para>
            /// </summary>
            protected void CalibrateWeightVectors()
            {
                if (this.MyKind != Kind.InputLayer)
                {
                    // この層の全ユニットに対して修正量を適用
                    for (int i = 0; i < this.units.Length; i++)
                        this.units[i].DoCorrect();
                }
                // ネストになっている次の層の修正量の適用を指示
                if (this.MyKind != Kind.OutputLayer) this.subsidiaryLayer.CalibrateWeightVectors();
                return;
            }
            /// <summary>
            /// 特徴ベクトルを渡して、演算結果を返す
            /// </summary>
            /// <param name="preLayerOutput">特徴ベクトル／前段の出力ベクトル</param>
            /// <param name="teacher">教師ベクトル</param>
            /// <returns>演算結果</returns>
            public Vector GetResult(Vector preLayerOutput, Vector teacher)
            {
                Vector myOutput = new Vector(this.units.Length);                    // ユニット出力格納用のベクトルを用意
                // --------入力層なら、出力値はそのまま入力値とする-----------------------------
                if (this.MyKind == Kind.InputLayer)
                {
                    myOutput = preLayerOutput;
                    for (int i = 0; i < this.units.Length; i++) this.units[i].Output = preLayerOutput[i];
                }
                else // ---他の層なら、この層の出力ベクトルを渡して計算させて答えを格納する---
                {
                    for (int i = 0; i < myOutput.Length; i++)
                        myOutput[i] = this.units[i].GetOutput(preLayerOutput);
                }
                // --------出力値を得る---------------------------------------------------------
                Vector result = new Vector(this.parameter.NumOfUnitInOutputLayer);  // 出力層の出力結果を格納するためにVectorクラス生成
                if (this.MyKind == Kind.OutputLayer)                                // 出力層での処理
                    result = myOutput;                                              // 出力層の出力はそのまま結果である
                else                                                                // 出力層以外での処理
                    result = this.subsidiaryLayer.GetResult(myOutput, teacher);     // 答えは、次の層から取得する（出力層のresultがリレーされる）
                // --------教師データがあれば結合係数の修正量を計算する-------------------------
                if (teacher != null && this.MyKind != Kind.InputLayer)              // 教師があれば、重みを修正する（入力層も実行してもいいけど、結合係数がないので意味ない。測度優先でここはスルーさせる。）
                {
                    this.CalcOwnUnitsError(teacher, myOutput);                      // 出力エラーを計算
                    this.CalibrateWeightAndThreshold(preLayerOutput, myOutput);     // 発火の閾値の調整および、結合係数の修正を実施する
                }
                // --------教師データがあり、この層が入力層であれば結合係数の修正を指示する-----
                if (teacher != null && this.MyKind == Kind.InputLayer)              // 最後にまとめて
                    this.CalibrateWeightVectors();                                  // 結合係数を調律する
                return result;
            }
            /// <summary>
            /// 重みベクトルを乱数で初期化する
            /// </summary>
            public void InitWeightVectorWithRandom()
            {
                if (this.MyKind != Kind.InputLayer)
                {
                    for (int i = 0; i < this.units.Length; i++) this.units[i].InitWeightVectorWithRandom();
                }
                if (this.MyKind != Kind.OutputLayer) this.subsidiaryLayer.InitWeightVectorWithRandom();
                return;
            }
            /// <summary>
            /// ファイル名を指定して結合係数を読み込ませる
            /// </summary>
            /// <param name="fname">ファイル名</param>
            public void Setup(string fname = "layer.ini")
            {
                if (System.IO.File.Exists(fname) == false) throw new SystemException("指定されたファイルは存在しません。");
                using (System.IO.StreamReader sr = new System.IO.StreamReader(fname, System.Text.Encoding.UTF8))
                {
                    int layerNum = -1;
                    int unitNum = -1;
                    Kind kind = Kind.NA;
                    string backupLine = "";
                    while (sr.EndOfStream == false)
                    {
                        string line = sr.ReadLine();                            // 一行読み込み
                        if (line.IndexOf("<Layer = ") >= 0)                     // 層番号を取得
                        { 
                            Regex ri = new Regex(@"\d+");
                            Match m = ri.Match(line);
                            layerNum = int.Parse(m.ToString());
                        }
                        else if (line.IndexOf("<unit num=") >= 0)               // ユニット番号を取得
                        { 
                            Regex ri = new Regex(@"\d+");
                            Match m = ri.Match(line);
                            unitNum = int.Parse(m.ToString());
                        }
                        else if (line.IndexOf("Hidden") >= 0) kind = Kind.HiddenLayer;  // 層の種類を取得
                        else if (line.IndexOf("Output") >= 0) kind = Kind.OutputLayer;
                        else if (layerNum == this.myLayerNum && this.MyKind == kind && backupLine == "<threshold>")
                        {
                            this.units[unitNum].Threshold = double.Parse(line);         // ユニットの閾値を取得
                        }
                        else if (layerNum == this.myLayerNum && this.MyKind == kind && backupLine == "<weight vector>")
                        {
                            Vector vect = new Vector(line);                             // 結合強度を取得
                            this.units[unitNum].SetWeightVector(vect);                  // 結合強度をセット
                        }
                        backupLine = line;
                    }
                }
                if (this.MyKind != Kind.OutputLayer) this.subsidiaryLayer.Setup(fname); // 下層レイヤもセット（出力層は下層がないので呼び出しはしない）
                return;
            }
            /// <summary>
            /// レイヤークラスのコンストラクタ
            /// <para>このクラスインスタンスを一つ生成すれば、入力層から出力層までを表現可能です。</para>
            /// <para>
            /// 本クラスを生成する際、numOfPreLayerUnitは0と設定することをお勧めします。
            /// 0でなくともエラーは出ませんが、入力層には結合係数が定義されないためです。
            /// </para>
            /// </summary>
            /// <param name="_parameter">ニューラルネットのパラメータ</param>
            /// <param name="_myLayerNum">層番号</param>
            /// <param name="numOfPreLayerUnit">前の層のユニット数（省略可能です。デフォルトでは0となります。）</param>
            public Layer(Parameter _parameter, int _myLayerNum, int numOfPreLayerUnit = 0)      // 3層以上と限定すれば、numOfPreLayerUnitは要らないんだけど2層も試したかったので導入した
            {
                this.parameter = _parameter;
                this.grossLayerNum = this.parameter.TotalLayer;
                this.myLayerNum = _myLayerNum;
                if (myRandom == null) myRandom = new Random();
                // ユニットの準備と、前層のユニット数を記憶する
                switch (this.MyKind)
                {
                    case Kind.InputLayer:                                                       // 入力層
                        this.units = new Unit[this.parameter.NumOfUnitInInputLayer];            // 層の種類に合わせて、ユニット数を宣言する
                        break;
                    case Kind.HiddenLayer:                                                      // 中間層
                        this.units = new Unit[this.parameter.NumOfUnitInHiddenLayer];
                        break;
                    case Kind.OutputLayer:                                                      // 出力層
                        this.units = new Unit[this.parameter.NumOfUnitInOutputLayer];
                        break;
                }
                for (int i = 0; i < this.units.Length; i++) this.units[i] = new Unit(numOfPreLayerUnit);        // 前の層のユニット数だけ結合係数を生成させる
                this.preLayerUnitNum = numOfPreLayerUnit;                                                       // 前の層のユニット数を記憶しておく
                if (this.MyKind != Kind.OutputLayer)
                    this.subsidiaryLayer = new Layer(this.parameter, this.myLayerNum - 1, this.units.Length);   // 出力層以外なら子レイヤを生成
            }
        }
    } // ニューラルネットワークのレイヤを構成
    /// <summary>
    /// ニューラルネットワーククラス本体
    /// </summary>
    public partial class NeuralNetwork
    {
        /************** メンバ変数 **********************/
        /// <summary>本クラスの構造を規定するパラメータ<para>中間層数などを格納する</para></summary>
        private Parameter parameter;
        /// <summary>演算処理を実行するレイヤークラス<para>再帰構造になっているので一つのオブジェクトで演算は完結している</para></summary>
        private Layer layer;
        /************** プロパティ **********************/
        /// <summary>
        /// 出力誤差の変化量
        /// </summary>
        public double VariationOfError { get { return this.layer.VariationOfError; } }
        /// <summary>
        /// 出力誤差の合計
        /// <para>絶対値をとった出力層における各ユニットの誤差を合計したものです。</para>
        /// </summary>
        public double TotalOutputError { get { return this.layer.TotalOutputError; } }
        /************** メソッド **********************/
        /// <summary>
        /// ニューラルネットワークの構成パラメータを返す
        /// </summary>
        /// <returns>ニューラルネットワークの構成パラメータ</returns>
        public Parameter GetParameter()
        {
            return this.parameter; 
        }
        /// <summary>
        /// ニューラルネットの構成パラメータ及び結合係数を文字列として出力する。
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder(10000);
            sb.Append(this.parameter.ToString());
            sb.Append(this.layer.ToString());
            return sb.ToString();
        }
        /// <summary>
        /// 指定ファイル名でニューラルネットの状態量をテキストファイルへ保存する
        /// <para>デフォルトのファイル名は「NN.ini」です。</para>
        /// </summary>
        /// <param name="fname">ファイル名</param>
        /// <param name="append">追記の有無<para>true：追記</para></param>
        public void Save(string fname = "NN.ini", Boolean append = true)
        {
            using (System.IO.StreamWriter sw = new System.IO.StreamWriter(fname, append, System.Text.Encoding.UTF8))
            {
                sw.WriteLine(this.ToString());
            }
            return;
        }
        /// <summary>
        /// 特徴ベクトルを基に識別を行う
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <returns>識別結果</returns>
        public double[] Recognize(double[] featureVector)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult(new Vector(featureVector), null);
            return result.ToDoubleArray();
        }
        /// <summary>
        /// 特徴ベクトルを基に識別を行う
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <returns>識別結果</returns>
        public Vector Recognize(Feature featureVector)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult((Vector)featureVector, null);
            return result;
        }
        /// <summary>
        /// 特徴データを基に識別を行う
        /// <para>入力ユニットが1つの場合に使用します。出力層の第0ユニットの出力を返します。関数近似の用途を想定しています。</para>
        /// </summary>
        /// <param name="feature">特徴データ</param>
        /// <returns>返値<para>出力層の第0ユニットの出力</para></returns>
        public double Recognize(double feature)
        {
            if (this.parameter.NumOfUnitInInputLayer != 1) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult(new Vector(feature), null);
            return result[0];
        }
        /// <summary>
        /// 特徴ベクトルを基に、学習を行う
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <param name="NeuralNetTeacher">教師ベクトル</param>
        /// <returns>識別結果</returns>
        public double[] Learn(double[] featureVector, double[] NeuralNetTeacher)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            if (this.parameter.NumOfUnitInOutputLayer != NeuralNetTeacher.Length) throw new SystemException("出力層のユニット数と教師ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult(new Vector(featureVector), new Vector(NeuralNetTeacher));
            return result.ToDoubleArray();
        }
        /// <summary>
        /// 特徴ベクトルを基に、学習を行う
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <param name="NeuralNetTeacher">教師ベクトル</param>
        /// <returns>識別結果</returns>
        public Vector Learn(Feature featureVector, Vector NeuralNetTeacher)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            if (this.parameter.NumOfUnitInOutputLayer != NeuralNetTeacher.Length) throw new SystemException("出力層のユニット数と教師ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult((Vector)featureVector, NeuralNetTeacher);
            return result;
        }
        /// <summary>
        /// 特徴ベクトルを基に、学習を行う
        /// <para>出力が一つしかない場合に特に便利です。例えばXORの学習時に利用してください。</para>
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <param name="NeuralNetTeacher">教師データ</param>
        /// <returns>演算結果</returns>
        public double Learn(Feature featureVector, double NeuralNetTeacher)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            if (this.parameter.NumOfUnitInOutputLayer != 1) throw new SystemException("出力層のユニット数と教師ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult((Vector)featureVector, new Vector(NeuralNetTeacher));
            return result[0];
        }
        /// <summary>
        /// 特徴ベクトルを基に、学習を行う
        /// <para>出力が一つしかない場合に特に便利です。例えばXORの学習時に利用してください。特徴ベクトルにはdouble型配列を用います。</para>
        /// </summary>
        /// <param name="featureVector">特徴ベクトル</param>
        /// <param name="NeuralNetTeacher">教師データ</param>
        /// <returns>演算結果</returns>
        public double Learn(double[] featureVector, double NeuralNetTeacher)
        {
            if (this.parameter.NumOfUnitInInputLayer != featureVector.Length) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            if (this.parameter.NumOfUnitInOutputLayer != 1) throw new SystemException("出力層のユニット数と教師ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult(new Vector(featureVector), new Vector(NeuralNetTeacher));
            return result[0];
        }
        /// <summary>
        /// 特徴データを基に、学習を行う
        /// <para>関数近似に利用可能です。</para>
        /// </summary>
        /// <param name="feature">特徴データ</param>
        /// <param name="NeuralNetTeacher">教師データ</param>
        /// <returns>演算結果</returns>
        public double Learn(double feature, double NeuralNetTeacher)
        {
            if (this.parameter.NumOfUnitInInputLayer != 1) throw new SystemException("入力層のユニット数と特徴ベクトルの長さが一致しません。");
            if (this.parameter.NumOfUnitInOutputLayer != 1) throw new SystemException("出力層のユニット数と教師ベクトルの長さが一致しません。");
            Vector result = this.layer.GetResult(new Vector(feature), new Vector(NeuralNetTeacher));
            return result[0];
        }
        /// <summary>
        /// ファイルを指定して結合係数を読み込ませる
        /// <para>ファイル名はデフォルトでは"NN.ini"です。</para>
        /// </summary>
        /// <param name="fname"></param>
        public void Setup(string fname = "NN.ini")
        {
            if (System.IO.File.Exists(fname) == false) throw new SystemException("指定されたファイルは存在しません。");
            this.layer.Setup(fname);                        // 実際の処理はlayerに丸投げ
            return;
        }
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="_parameter">ニューラルネットワーク構造を規定するパラメータ</param>
        public NeuralNetwork(Parameter _parameter)
        {
            this.parameter = _parameter;
            this.layer = new Layer(this.parameter, this.parameter.TotalLayer - 1);
            this.layer.LearningCoefficient = this.parameter.LearningCoefficient;
            return;
        }
    } // ニューラルネットワーククラス本体
    /// <summary>
    /// モデルのクラス
    /// </summary>
    public class ClassModel
    {
        /* メンバ変数 ***********************************/
        /// <summary>クラス名</summary>
        private readonly string className;
        /// <summary>クラス構成メンバ</summary>
        private Feature[] members;
        /// <summary>構成メンバの大きさ<para>追加される特徴ベクトルのサイズチェック用です。</para></summary>
        private int memberSize = 0;
        /// <summary>特徴ベクトルの読み出しポイント<para>GetFeatureメソッドの読み出し時に使用する。</para></summary>
        private int readPoint = 0;
        /* プロパティ ***********************************/
        /// <summary>
        /// メンバ数
        /// </summary>
        public int Length { get { return this.members.Length; } }
        /// <summary>
        /// クラス名（ID）
        /// </summary>
        public string ID { get { return this.className; } }
        /* インデクサ ***********************************/
        /// <summary>
        /// インデクサ
        /// <para>本モデルが保持している特徴ベクトルにアクセスします。</para>
        /// </summary>
        /// <param name="i">インデックス番号</param>
        /// <returns>指定インデックスの特徴ベクトル</returns>
        public Feature this[int i]
        {
            get{
                if (i < 0 || i >= this.Length) throw new SystemException("インデックスの指定範囲が不正です。");
                return this.members[i];
            }
        }
        /* メソッド *************************************/
        /// <summary>
        /// クラスの構成メンバ（特徴ベクトル）を文字列化する
        /// <para>統計処理に利用可能なように最後の要素にクラス名をラベルとして追加しています。</para>
        /// </summary>
        /// <returns>文字列化したクラスメンバ</returns>
        public override string ToString()
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder(1500);
            for (int i = 0; i < this.members.Length; i++) sb.Append(this.members[i].ToString()).Append(this.ID).Append("\n");
            return sb.ToString();
        }
        /// <summary>
        /// 指定された要素番号の構成メンバを返す
        /// </summary>
        /// <param name="index">要素番号</param>
        /// <returns>指定された構成メンバ（特徴ベクトル）</returns>
        public Feature GetMember(int index)
        {
            return this.members[index];
        }
        /// <summary>
        /// 保持する特徴ベクトルを一つ返す
        /// <para>特徴ベクトルをリングバッファとして取り扱い、本メソッドを呼び出すたびに読み込み位置を一つ進める。</para>
        /// </summary>
        /// <returns>特徴ベクトル</returns>
        public Feature GetFeature()
        {
            Feature ans = this[this.readPoint++];
            this.readPoint %= this.Length;
            return ans;
        }
        /// <summary>
        /// 特徴ベクトルの追加
        /// </summary>
        /// <param name="newMember">追加メンバ</param>
        public void Add(Feature newMember)
        {
            if (this.members == null) this.members = new Feature[0];
            Array.Resize<Feature>(ref this.members, this.members.Length + 1);
            this.members[this.members.Length - 1] = newMember;
            if (this.members.Length == 1)
                this.memberSize = newMember.Length;
            else
                if (this.memberSize != newMember.Length)throw new SystemException("既に登録された特徴ベクトルのサイズと引数のサイズが異なります。");
            return;
        }
        // コンストラクタ関係------------------------------------------------------------------
        /// <summary>
        /// コンストラクタ
        /// <para>クラスIDのみを設定するコンストラクタです。メンバーは要素数0で初期化されます。</para>
        /// </summary>
        /// <param name="_id">クラス名</param>
        public ClassModel(string _id)
        {
            this.className = _id;
            this.members = new Feature[0];
        }
        /// <summary>
        /// メンバの初期化付コンストラクタ
        /// <para>引数で渡した配列は要素を一つ一つコピーするので割と安心です。</para>
        /// </summary>
        /// <param name="_id">クラス名</param>
        /// <param name="_members">メンバ</param>
        /// <exception cref="SystemException">メンバがnullだとスロー</exception>
        public ClassModel(string _id, Feature[] _members)
        {
            this.className = _id;
            if (_members == null) throw new SystemException("メンバがnullです。");
            if (_members.Length == 0) throw new SystemException("配列サイズが0です。");
            this.members = new Feature[_members.Length];
            for (int i = 0; i < this.members.Length; i++) this.members[i] = _members[i];
            this.memberSize = _members[0].Length;
        }
    }
    /// <summary>
    /// ニューラルネットワークの教師クラス
    /// <para>
    /// ニューラルネットワークを教育するクラスです。
    /// 使用するには、インスタンス生成時にニューラルネットのパラメータを設定して下さい。
    /// [今後の予定]
    /// 2011/12/24 ジャックナイフ判定法の導入は必要だろうなぁ。
    /// </para>
    /// </summary>
    public class NeuralNetTeacher
    {
        /* メンバ変数 *********************************************************/
        /// <summary>クラスモデル</summary>
        private ClassModel[] _class;
        /// <summary>ニューラルネットワーク</summary>
        private NeuralNetwork _NNet;
        /// <summary>ニューラルネットワーク用パラメータ</summary>
        private NeuralNetwork.Parameter _NNparameter;
        /// <summary>中間層のユニット数</summary>
        private readonly int numOfUnitInHiddenLayer;
        /// <summary>中間層数</summary>
        private readonly int numOfHiddenLayer;
        /// <summary>学習係数</summary>
        private readonly double learningCoefficient;
        /* プロパティ *********************************************************/
        /// <summary>
        /// クラスモデル数
        /// </summary>
        public int Length { get { return this._class.Length; } }
        /// <summary>
        /// 登録クラス内に同じIDを持つものがないかチェックし、在ればtrueを返す
        /// </summary>
        private Boolean ExistOfSameID
        {
            get {
                Hashtable hoge = new Hashtable();
                for (int i = 0; i < this._class.Length; i++)
                {
                    if (hoge.Contains(this._class[i].ID) == false)
                        hoge.Add(this._class[i].ID, null);
                    else
                        return true;
                }
                return false;
            }
        }
        /// <summary>
        /// 登録されれている全特徴ベクトルについて2重登録を確認し、在ればtrueを返す
        /// </summary>
        private Boolean ExistOfSameFeature
        {
            get {
                Hashtable hoge = new Hashtable();
                for (int i = 0; i < this._class.Length; i++)
                {
                    for (int k = 0; k < this._class[i].Length; k++) 
                    {
                        if (hoge.Contains(this._class[i].GetMember(k).GetHashCode()) == false)
                            hoge.Add(this._class[i].GetMember(k).GetHashCode(), null);
                        else
                            return true;
                    }
                }
                return false;
            }
        }
        /// <summary>
        /// 保持しているクラスモデルのID
        /// </summary>
        public string[] ModelIDs
        {
            get { 
                string[] id = new string[this._class.Length];
                for (int i = 0; i < id.Length; i++) id[i] = this._class[i].ID;
                return id;
            }
        }
        /// <summary>
        /// 学習準備
        /// <para>学習準備が整っていればtrue</para>
        /// </summary>
        public Boolean SetupComplete
        {
            get {
                if (this._class == null)
                    return false;
                else
                    return true;
            }
        }
        /// <summary>
        /// モデル個々の識別率
        /// <para>識別率はモデルIDとセットにしてハッシュテーブルで表します。</para>
        /// <para>なお、処理の実態はGetAccurecyRate()の出力値です。</para>
        /// <para>格納される値の定義：モデルの正答数/モデルの特徴ベクトル数</para>
        /// <para>演算が不可能な場合はnullが返ります。</para>
        /// <para>演算には時間が少々かかりますので、まとまった学習を実施後にチェックするようにして下さい。</para>
        /// </summary>
        public Hashtable EachAccurecyRate
        {
            get {
                if (this.SetupComplete == false)
                    return null;
                else
                    return this.GetEachAccurecyRate();
            }
        }
        /// <summary>
        /// 全モデルの総合識別率
        /// <para>処理の実態はGetTotalAccurecyRate()の出力値です。</para>
        /// <para>格納される値の定義：正答の総数/特徴ベクトルの総数</para>
        /// <para>演算が不可能な場合はnullが返ります。</para>
        /// <para>演算には時間が少々かかりますので、まとまった学習を実施後にチェックするようにして下さい。</para>
        /// </summary>
        public double? TotalAccurecyRate
        {
            get
            {
                if (this.SetupComplete == false)
                    return null;
                else
                    return this.GetTotalAccurecyRate();
            }
        }
        /// <summary>
        /// 学習に使用するニューラルネットワークの構成情報
        /// </summary>
        public NeuralNetwork.Parameter Parameter { get { return this._NNparameter; } }
        /* メソッド **************************************************************/
        /// <summary>
        /// 学習に使用しているニューラルネットワーククラスオブジェクトを返す
        /// <para>学習結果を利用して識別を行う際に使用することを想定しています。</para>
        /// </summary>
        /// <returns>本クラスが保持するニューラルネットワーククラスオブジェクト</returns>
        public NeuralNetwork GetNeuralNetwork()
        {
            return this._NNet;
        }
        /// <summary>
        /// 登録されているクラスの特徴ベクトルをファイルへ出力する
        /// <para>モデル毎にファイルを分割して出力します。</para>
        /// <para>ファイル名にはモデルのIDが使用されます。</para>
        /// <para>ファイルは上書きされます。</para>
        /// <para>文字コードはUTF8です。</para>
        /// </summary>
        public void SaveModelFeatures()
        {
            for (int i = 0; i < this._class.Length; i++)
            {
                string fname = this._class[i].ID + ".fea";
                System.IO.File.WriteAllText(fname, this._class[i].ToString(), Encoding.UTF8);     //ファイルが存在しているときは、上書きする
            }
            return;
        }
        /// <summary>
        /// モデル名一覧、ニューラルネットワークのパラメータ、学習した結合係数を保存する（学習していなくても保存できるけど）
        /// <para>ファイルは上書きされます。</para>
        /// </summary>
        /// <param name="fname">ファイル名<para>デフォルトはDiscriminatorクラスのSetup()に合わせて"discriminator.ini"です。</para></param>
        public void SaveLearnigOutcome(string fname = "discriminator.ini")
        {
            try
            {
                // モデルIDを保存
                string str = "<ID>\n";
                for (int i = 0; i < this.Length; i++)
                {
                    str = str + this.ModelIDs[i];
                    if (i < this.Length - 1) str = str + ",";
                }
                str = str + "\n</ID>\n";
                System.IO.File.WriteAllText(fname, str, Encoding.UTF8);     //ファイルが存在しているときは、上書きする
                // 以下、追記でニューラルネットワークのパラメータ，結合係数の順で追記保存する。
                this._NNet.Save(fname, true);
            }
            catch
            {
                MessageBox.Show("結合係数の保存中にエラーが発生しました。", "エラー");
            }
            return;
        }
        /// <summary>
        /// 指定されたクラス名に応じた教師ベクトルを返す
        /// </summary>
        /// <param name="id">クラス名</param>
        /// <returns>教師ベクトル</returns>
        private Vector CreateTeachingVector(string id)
        {
            Vector teachingVector = new Vector(this._class.Length);
            for (int i = 0; i < this._class.Length; i++)
            {
                if (id == this._class[i].ID)
                    teachingVector[i] = 1.0;
                else
                    teachingVector[i] = 0.0;
            }
            return teachingVector;
        }
        /// <summary>
        /// 指定されたクラス番号に応じた教師ベクトルを返す
        /// </summary>
        /// <param name="index">クラス要素番号</param>
        /// <returns>教師ベクトル</returns>
        private Vector CreateTeachingVector(int index) 
        {
            Vector teachingVector = new Vector(this._class.Length);
            for (int i = 0; i < this._class.Length; i++)
            {
                if (i == index)
                    teachingVector[i] = 1.0;
                else
                    teachingVector[i] = 0.0;
            }
            return teachingVector;
        }
        /// <summary>
        /// クラスのIDを文字列配列で返す
        /// </summary>
        /// <returns>ID名を配列にしたもの</returns>
        public string[] GetIDnames()
        { 
            string[] ans = new string[this._class.Length];
            for (int i = 0; i < this._class.Length; i++)
            {
                ans[i] = this._class[i].ID;
            }
            return ans;
        }
        /// <summary>
        /// クラスモデルを追加する
        /// </summary>
        /// <param name="newClass">追加するクラスモデル</param>
        /// <exception cref="SystemException">モデルの2重登録があるとスロー</exception>
        public void AddModel(ClassModel newClass)
        {
            if (this._class == null) this._class = new ClassModel[0];
            Array.Resize(ref this._class, this._class.Length);
            this._class[this._class.Length - 1] = newClass;
            if (this.ExistOfSameID) throw new SystemException("モデルの2重登録が行われました。");
            return;
        }
        #region // 判別表など、成績評価関係----------------------------------------------------------
        /// <summary>
        /// 一通りの判別を行い、結果をヒストグラム（表：実態はVector[]）にまとめる
        /// </summary>
        /// <returns>判別結果を表にまとめたもの（実態はVector[]）</returns>
        /// <exception cref="SystemException">クラスモデルがセットされた状態でなければ例外をスロー</exception>
        private Vector[] CreateDecisionChart()
        {
            if (this.SetupComplete == false) throw new SystemException("実験の準備はまだ完了しておりません。モデルを追加して下さい。");
            // まずは判定
            Vector[] decisionChart = new Vector[this.Length];                   // 判定結果を格納する
            for (int modelIndex = 0; modelIndex < this.Length; modelIndex++)    // 保持しているモデルの数だけループ
            {
                decisionChart[modelIndex] = new Vector(this.Length);
                // 特徴モデルを順番に呼び出す
                for (int featureIndex = 0; featureIndex < this._class[modelIndex].Length; featureIndex++)
                {
                    Feature feature = this._class[modelIndex][featureIndex];    // 特徴ベクトルをモデルから取得
                    Vector outputForDebug = this._NNet.Recognize(feature);      // ニューラルネットの演算
                    // 最大値となるインデックスを検索
                    double max = double.MinValue;
                    int maxIndex = 0;
                    for (int j = 0; j < outputForDebug.Length; j++)
                    {
                        if (max < outputForDebug[j])
                        {
                            max = outputForDebug[j];
                            maxIndex = j;
                        }
                    }
                    // 最大値となったインデックスについて、加算
                    decisionChart[modelIndex][maxIndex] += 1;
                }
            }
            return decisionChart;
        }
        /// <summary>
        /// 個別のモデルの識別率を返す
        /// <para>格納される値の定義：モデルの正答数/モデルの特徴ベクトル数</para>
        /// <para>クラスモデルがセットされた状態でなければ例外をスローします。</para>
        /// </summary>
        /// <returns>クラスモデルIDと識別率をハッシュテーブルにまとめたもの</returns>
        /// <exception cref="SystemException">クラスモデルがセットされた状態でなければ例外をスロー</exception>
        public Hashtable GetEachAccurecyRate()
        {
            if (this.SetupComplete == false) throw new SystemException("準備はまだ完了しておりません。モデルを追加して下さい。");
            Hashtable ans = new Hashtable();
            // まずは判定
            Vector[] decisionChart = this.CreateDecisionChart();                   // 判定結果を格納する
            // 成績を取りまとめ
            for (int modelIndex = 0; modelIndex < this.Length; modelIndex++)
            {
                double accuracyRate = decisionChart[modelIndex][modelIndex] / this._class[modelIndex].Length;
                ans.Add(this._class[modelIndex].ID, accuracyRate);
            }
            return ans;
        }
        /// <summary>
        /// 全モデルの総合での識別率を返す
        /// <para>本メソッドが返すTotal Accuracy Rateの定義：Total Accuracy Rate = 正答の総数/特徴ベクトルの総数</para>
        /// <para>Total Accuracy Rateの定義からして、正答率の高いモデルの特徴ベクトルが多ければ全体の正答率も高くなることに留意して下さい。</para>
        /// <para>クラスモデルがセットされた状態でなければ例外をスローします。</para>
        /// </summary>
        /// <returns>総合識別率</returns>
        /// <exception cref="SystemException">クラスモデルがセットされた状態でなければ例外をスロー</exception>
        public double GetTotalAccurecyRate()
        {
            if (this.SetupComplete == false) throw new SystemException("準備はまだ完了しておりません。モデルを追加して下さい。");
            // まずは判定
            Vector[] decisionChart = this.CreateDecisionChart();                    // 判定結果を格納する
            // 成績を取りまとめ
            double featureSize = 0.0;                                               // 特徴ベクトルの総数
            double rightAnswerSize = 0.0;                                           // 正答の総数
            for (int modelIndex = 0; modelIndex < this.Length; modelIndex++)
            {
                featureSize += this._class[modelIndex].Length;
                rightAnswerSize += decisionChart[modelIndex][modelIndex];
            }
            return rightAnswerSize / featureSize;
        }
        /// <summary>
        /// 判定表をCSV形式で保存する
        /// </summary>
        /// <param name="fname">保存ファイル名</param>
        /// <exception cref="SystemException">クラスモデルがセットされた状態でなければ例外をスロー</exception>
        public void SaveDecisionChart(string fname = "decision chart.csv")
        {
            if (this.SetupComplete == false) throw new SystemException("実験の準備はまだ完了しておりません。モデルを追加して下さい。");
            // まずは判定
            Vector[] decisionChart = this.CreateDecisionChart();                   // 判定結果を格納する
            // 次はファイルへの出力
            using (System.IO.StreamWriter sw = new System.IO.StreamWriter(fname, false, Encoding.UTF8))
            {
                // ラベルを出力
                for (int j = 0; j < this.Length; j++) sw.Write("," + this.ModelIDs[j].ToString());
                sw.Write("\n");
                // 識別結果を出力
                for (int i = 0; i < decisionChart.Length; i++)
                {
                    sw.Write(this.ModelIDs[i].ToString());
                    sw.WriteLine("," + decisionChart[i].ToString("0"));
                }
                // 正答率を出力（あるモデルがそのモデルであると判別された割合）
                Hashtable eachAccuracyRate = this.GetEachAccurecyRate();
                double totalAccuracyRate = this.GetTotalAccurecyRate();
                sw.Write("Accuracy Rate");
                for (int i = 0; i < eachAccuracyRate.Count; i++)
                {
                    sw.Write("," + ((double)(eachAccuracyRate[this._class[i].ID])).ToString("0.00"));
                }
                sw.Write("\n");
                sw.WriteLine("Total Accuracy Rate," + totalAccuracyRate.ToString("0.00"));
            }
            return;
        }
        #endregion
        #region // 学習関係--------------------------------------------------------------------------
        /// <summary>
        /// 学習を実行させる
        /// <para>学習強度は1:1に設定されますので、各モデルの保持する特徴ベクトルは等しい回数呼び出される。</para>
        /// </summary>
        /// <param name="learningTimes">学習回数</param>
        public void Learn(int learningTimes)
        { 
            int[] learningStrength = new int[this.Length];
            for(int i = 0; i < learningStrength.Length; i++)learningStrength[i] = 1;
            this.Learn(learningTimes, learningStrength);
        }
        /// <summary>
        /// 学習強度を格納した配列のチェックを行う
        /// <para>全て0以下であれば学習する必要がないのでfalseを返す。</para>
        /// </summary>
        /// <param name="learningStrength">学習強度</param>
        /// <returns>学習させる必要があるか<para>true：必要あり</para></returns>
        private Boolean CheckLearningStrength(int[] learningStrength)
        {
            Boolean ans = false;
            for (int i = 0; i < learningStrength.Length; i++)
                if (learningStrength[i] > 0) ans = true;
            return ans;
        }
        /// <summary>
        /// 学習を実行させる
        /// <para>学習強度付</para>
        /// </summary>
        /// <param name="learningTimes">学習回数</param>
        /// <param name="learningStrength">
        /// 学習強度（要素数はモデル数と一致すること）
        /// <para>
        /// 学習強度が各モデルの特徴数に比例しないように導入しました。
        /// 0に設定すると学習は行われません。
        /// </para>
        /// </param>
        public void Learn(int learningTimes, int[] learningStrength)
        {
            if (this.SetupComplete == false) throw new SystemException("学習の準備はまだ完了しておりません。モデルを追加して下さい。");
            if (this.Length != learningStrength.Length) throw new SystemException("学習強度の要素数がモデル数と一致しません。");// モデル数と一致していることをチェック
            // 学習をスタート
            int learningCount = 0;
            double[] errorEachModel = new double[this.Length];
            Boolean need = CheckLearningStrength(learningStrength);                     // 学習させる必要性を学習強度からチェック
            while (learningCount < learningTimes && need)                               // 学習回数をチェックしながらループ
            {
                // 保持しているモデルの数だけループ
                for (int modelIndex = 0; modelIndex < this.Length; modelIndex++)
                {
                    Vector teature = this.CreateTeachingVector(modelIndex);
                    double error = 0.0;
                    int count = 0;
                    // 指定された強度分配に基づき、計算した回数分モデルを呼び出す
                    for (int i = 0; i < learningStrength[modelIndex]; i++)
                    {
                        Feature feature = this._class[modelIndex].GetFeature();         // 特徴ベクトルをモデルから取得
                        Vector outputForDebug = this._NNet.Learn(feature, teature);     // ニューラルネットワークの学習
                        learningCount++;
                        error += this._NNet.TotalOutputError;
                        count++;
                    }
                    error /= (double)count;                                             // 特徴量1件当たりの誤差量を計算
                    errorEachModel[modelIndex] = error;                                 // シーケンス上では何の役にも立っていないが、デバッグ用と思ってください。将来的には誤差の収束条件を入れたいと思っています。
                }
            }
            return;
        }
        #endregion
        #region // コンストラクタ・初期化関係 -------------------------------------------------------
        /// <summary>
        /// ファイル名からクラスモデル名を抽出して返す
        /// </summary>
        /// <param name="fname">ファイル名</param>
        /// <returns>クラスモデル名</returns>
        private string GetModelIDname(string fname)
        {
            string name = System.IO.Path.GetFileNameWithoutExtension(fname);
            string[] field = name.Split('_');
            return field[0];
        }
        /// <summary>
        /// *.feaで該当したファイル名の一覧から、クラスモデル名を抽出して一覧として返す
        /// </summary>
        /// <param name="fnames">*.feaで該当したファイル名の一覧</param>
        /// <returns>クラスモデル名の一覧</returns>
        private Hashtable GetModelID(string[] fnames)
        {
            Hashtable ans = new Hashtable();
            int modelIndex = 0;

            for (int i = 0; i < fnames.Length; i++)
            {
                string id = this.GetModelIDname(fnames[i]);
                if (ans.Contains(id) == false)
                {
                    ans.Add(id, modelIndex);
                    modelIndex++;
                }
            }
            return ans;
        }
        /// <summary>
        /// フォルダを指定してモデルデータを読み込む
        /// <para>読み込んだファイル名の一覧よりモデルリストを作成し、ファイルの中身から入力層に必要なユニット数を確保します。</para>
        /// <para>読み込まれた特徴量がモデルによって異なる場合は、学習の段階で入力ユニット数と特徴ベクトルのサイズの不一致に関するエラーがスローされます。</para>
        /// </summary>
        /// <param name="dirName">特徴データの格納されたフォルダ名</param>
        /// <exception cref="SystemException">ファイルから読み込んだ特徴ベクトルの長さとニューラルネットの入力層ユニット数が一致しない場合にスロー</exception>
        public void Setup(string dirName)
        {
            if (System.IO.Directory.Exists(dirName) == false) throw new SystemException("指定されたフォルダは存在しません。");
            string[] fnames = System.IO.Directory.GetFiles(dirName, "*.fea");   // フォルダ内の指定拡張子を持つファイル一覧を取得
            Hashtable modelID = this.GetModelID(fnames);                        // ファイル名の一覧からユニークなモデルIDを取得
            this._class = new ClassModel[modelID.Count];                        // 存在するファイルの数だけモデルを生成する
            int featureSize = 0;
            // クラスモデルを構成
            for (int i = 0; i < fnames.Length; i++)
            {
                string id = this.GetModelIDname(fnames[i]);                     // ファイル名からモデル名を取得
                if (this._class[(int)modelID[id]] == null) this._class[(int)modelID[id]] = new ClassModel(id); // インスタンスが未生成なら作る
                using (System.IO.StreamReader sr = new System.IO.StreamReader(fnames[i], System.Text.Encoding.GetEncoding("shift_jis")))
                {
                    try
                    {
                        while (sr.EndOfStream == false)
                        {
                            string line = sr.ReadLine();                            // 一行読み込み
                            Feature feature = new Feature(line);                    // 読み込んだテキストデータから特徴ベクトルを生成
                            this._class[(int)modelID[id]].Add(feature);             // 特徴を追加
                            featureSize = feature.Length;                           // 特徴ベクトルの要素数を取得
                        }
                    }
                    catch (Exception e)
                    {
                        string message = "エラーがスローされました。\nエラーメッセージ：" + e.Message + "エラーの発生したファイル名：" + fnames[i];
                        Console.WriteLine(message);
                        throw new SystemException(message);                         // さらにスロー
                    }
                }
            }
            // ニューラルネットを構成
            this._NNparameter = new NeuralNetwork.Parameter(featureSize, this.numOfUnitInHiddenLayer, modelID.Count, this.numOfHiddenLayer, this.learningCoefficient);  // ニューラルネットのパラメータを復元
            this._NNet = new NeuralNetwork(this._NNparameter);                  // ニューラルネットクラスのインスタンスを生成
            return;
        }
        /* コンストラクタ ----------------------------------------------------------*/
        /// <summary>
        /// NeuralNetTeacherクラスのコンストラクタ
        /// <para>ニューラルネットのパラメータを引数に設定して下さい。</para>
        /// </summary>
        /// <param name="_numOfUnitInHiddenLayer">中間層のユニット数<para>この数が多いと、過学習に陥りやすくなります。n個で2^n状態を表すことができることを考慮して決定して下さい。</para></param>
        /// <param name="_numOfHiddenLayer">中間層数<para>この数+1だけ識別面（超平面）が形成されます。特徴量の分布をよく検討しなければまずいことになりそうです。</para></param>
        /// <param name="_learningCoefficient">学習係数<para>この数が大きいと結合係数の修正量が大きくなりすぎていつまでも収束しなくなります。かといって小さいといつまでも終わりません。カットアンドトライで確認してみてください。</para></param>
        public NeuralNetTeacher(int _numOfUnitInHiddenLayer, int _numOfHiddenLayer, double _learningCoefficient)
        {
            this.numOfUnitInHiddenLayer = _numOfUnitInHiddenLayer;
            this.numOfHiddenLayer = _numOfHiddenLayer;
            this.learningCoefficient = _learningCoefficient;
            return;
        }
        #endregion
    }
    /// <summary>
    /// 識別クラス
    /// <para>ニューラルネットワークを利用した識別を行う。</para>
    /// </summary>
    public class Discriminator
    {
        /* 構造体 ********************************************************************/
        /// <summary>
        /// 識別結果を簡単に表現するための構造体
        /// <para>クラス名を表すIDと、尤度を格納する。</para>
        /// </summary>
        public struct IDandLikelihood
        {
            /* メンバ変数 ***********************************/
            /// <summary>ID名</summary>
            private readonly string id;
            /// <summary>尤度</summary>
            private readonly double likelihood;
            /* プロパティ ***********************************/
            /// <summary>
            /// ID名
            /// </summary>
            public string ID { get { return this.id; } }
            /// <summary>
            /// 尤度
            /// </summary>
            public double Likelihood { get { return this.likelihood; } }
            /* メソッド ***********************************/
            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="_id">ID名</param>
            /// <param name="_likelihood">尤度</param>
            public IDandLikelihood(string _id, double _likelihood)
            {
                this.id = _id;
                this.likelihood = _likelihood;
            }
        }
        /* メンバ変数 ********************************************************************/
        /// <summary>ニューラルネットクラス</summary>
        private NeuralNetwork _NNet;
        /// <summary>ニューラルネットのパラメータ</summary>
        private NeuralNetwork.Parameter _NNparameter;
        /// <summary>ID名</summary>
        private string[] _IDnames;
        /* プロパティ ********************************************************************/
        /// <summary>
        /// 利用準備状況
        /// </summary>
        public Boolean Ready 
        {
            get {
                if (this._IDnames != null && this._NNet != null)
                    return true;
                else
                    return false;
            }
        }
        /* メソッド **********************************************************************/
        /// <summary>
        /// 識別用ニューラルネットの設定ファイルを読み込む
        /// <para>デフォルトのファイル名は"discriminator.ini"です。</para>
        /// </summary>
        /// <param name="fname">ファイル名</param>
        public void Setup(string fname = "discriminator.ini")
        {
            if (System.IO.File.Exists(fname) == false) throw new SystemException("指定されたファイルは存在しません。");
            try
            {
                // モデルID名をファイルより読み込む
                using (System.IO.StreamReader sr = new System.IO.StreamReader(fname, System.Text.Encoding.UTF8))
                {
                    Boolean idFound = false;
                    while (sr.EndOfStream == false)
                    {
                        string line = sr.ReadLine();                            // 一行読み込み
                        if (idFound)
                        {
                            string[] field = line.Split(',');
                            this._IDnames = new string[field.Length];
                            for (int i = 0; i < this._IDnames.Length; i++) this._IDnames[i] = field[i];
                            idFound = false;
                        }
                        if (line.IndexOf("ID") >= 0 && line.IndexOf("/ID") < 0) idFound = true;
                    }
                }
                // パラメータ復元
                this._NNparameter = new NeuralNetwork.Parameter(fname);
                // ニューラルネット復元
                this._NNet = new NeuralNetwork(this._NNparameter);
                this._NNet.Setup(fname);
            }
            catch
            {
                MessageBox.Show("DiscriminatorクラスのSetup()にてファイル読み込み時にエラーが発生しました。", "エラー");
            }
            return;
        }
        /// <summary>
        /// ニューラルネットの演算結果を分かり易くIDと尤度にまとめて返す
        /// </summary>
        /// <param name="outputVector">ニューラルネットの出力ベクトル</param>
        /// <param name="productionVolume">配列の生成数（1以上かつ出力ベクトルの次元を超えないようにして下さい）</param>
        /// <returns>ID・尤度の組を配列に加工したもの<para>本クラスのIDnamesメンバがセットされていない場合は""が返ります。</para></returns>
        /// <exception cref="SystemException">要求サイズに異常があるとスロー</exception>
        /// <example>
        /// 以下にサンプルコードを示します。
        /// このコードは処理の流れを大まかに示しています。
        /// 実際にはnnclassは結合係数等をセットしなければ識別は行えません。
        /// <code>
        /// NeuralNetwork.Parameter para = new NeuralNetwork.Parameter(2, 6, 1, 1, 0.01);           // ニューラルネットのパラメータ設定
        /// NeuralNetwork nnclass = new NeuralNetwork(para);                                        // ニューラルネットクラスのインスタンス生成
        /// double[] feature10 = new double[2] { 1, 0 };                                            // 特徴ベクトル生成（XORを意識している）
        /// 
        /// IDandLikelihood[] result = nnclass.ToIDandLikelihood(nnclass.Recognize(feature10), 3);  // 出力ベクトルをリレーさせて、構造体配列に変換している
        /// </code>
        /// </example>
        private IDandLikelihood[] GetIDandLikelihood(Vector outputVector, int productionVolume)
        {
            IDandLikelihood[] ans = new IDandLikelihood[productionVolume];
            double val = double.MaxValue;

            if (productionVolume == 0 || productionVolume > outputVector.Length) throw new SystemException("要求サイズが異常です。");
            for (int i = 0; i < productionVolume; i++)  // 大きい順に並び替えながら要素番号も把握する
            { 
                double max = 0.0;
                int index = -1;
                for (int k = 0; k < outputVector.Length; k++)
                {
                    if (max < outputVector[k] && val > outputVector[k])
                    {
                        max = outputVector[k];
                        index = k;
                    }
                }
                val = max;
                if (this._IDnames != null && index != -1)
                    ans[i] = new IDandLikelihood(this._IDnames[index], max);
                else
                    ans[i] = new IDandLikelihood("Non", max);
            }
            return ans;
        }
        /// <summary>
        /// 識別結果をIDと尤度で返す
        /// </summary>
        /// <param name="feature">特徴ベクトル</param>
        /// <returns>IDと尤度をセットにした識別結果</returns>
        public IDandLikelihood Discriminate(Feature feature)
        {
            IDandLikelihood ans = new IDandLikelihood();
            if (this.Ready)
            {
                Vector result = this._NNet.Recognize(feature);
                ans = this.GetIDandLikelihood(result, 1)[0];
            }
            return ans;
        }
        /// <summary>
        /// 識別結果をIDと尤度の配列で返す
        /// </summary>
        /// <param name="feature">特徴ベクトル</param>
        /// <param name="productionVolume">配列の生成数（1以上かつ出力ベクトルの次元を超えないようにして下さい）</param>
        /// <returns>IDと尤度をセットにした識別結果の配列</returns>
        public IDandLikelihood[] Discriminate(Feature feature, int productionVolume)
        {
            IDandLikelihood[] ans = null;
            if (this.Ready)
            {
                Vector result = this._NNet.Recognize(feature);
                ans = this.GetIDandLikelihood(result, productionVolume);
            }
            return ans;
        }
        /// <summary>
        /// ニューラルネットのインスタンスとモデル名のリストを渡して初期化するコンストラクタ
        /// </summary>
        /// <param name="NNet">ニューラルネットワーククラスオブジェクト</param>
        /// <param name="IDnames">モデルIDのリスト<para>リストの要素番号とモデル名は関係があるので適当な順序で渡さない様に。あくまで学習器(NeuralNetTeacher)から取得して下さい。</para></param>
        /// <exception cref="SystemException">ニューラルネットのインスタンスが確保されていない場合と、モデルIDの数と出力層ユニット数が一致しなければスロー</exception>
        public Discriminator(NeuralNetwork NNet, string[] IDnames)
        {
            if (NNet == null) throw new SystemException("ニューラルネットのインスタンスが確保されていません。");
            this._IDnames = new string[IDnames.Length];
            for (int i = 0; i < IDnames.Length; i++) this._IDnames[i] = IDnames[i];
            this._NNet = NNet;
            this._NNparameter = this._NNet.GetParameter();
            if (this._NNparameter.NumOfUnitInOutputLayer != this._IDnames.Length) throw new SystemException("モデルIDの数と出力層ユニット数が一致しません。");
        }
        /// <summary>
        /// 引数なしのコンストラクタ
        /// <para>Setup()を呼び出して初期化して下さい。</para>
        /// </summary>
        public Discriminator()
        { 
        
        }
    }

}
